@@ -80,6 +80,7 @@
 #include "dhcp.h"
 #include "if.h"
 #include "ipv4.h"
+#include "ipv4ll.h"
 #include "ipv6.h"
 #include "ipv6nd.h"
 
@@ -242,35 +243,19 @@ if_conf(struct interface *ifp)
 
 /* XXX work out Virtal Interface Masters */
 int
-if_vimaster(__unused const char *ifname)
+if_vimaster(__unused const struct dhcpcd_ctx *ctx, __unused const char *ifname)
 {
 
 	return 0;
 }
 
 static int
-_open_link_socket(struct sockaddr_nl *nl, int flags, int protocol)
+_open_link_socket(struct sockaddr_nl *nl, int protocol)
 {
 	int fd;
 
-#ifdef SOCK_CLOEXEC
-	if (flags)
-		flags = SOCK_CLOEXEC;
-	fd = socket(AF_NETLINK, SOCK_RAW | flags, protocol);
-	if (fd == -1)
+	if ((fd = xsocket(AF_NETLINK, SOCK_RAW, protocol, O_CLOEXEC)) == -1)
 		return -1;
-#else
-	fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
-	if (fd == -1)
-		return -1;
-	if (flags &&
-	    (flags = fcntl(fd, F_GETFD, 0)) == -1 ||
-	    fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1)
-	{
-		close(fd);
-	        return -1;
-	}
-#endif
 	nl->nl_family = AF_NETLINK;
 	if (bind(fd, (struct sockaddr *)nl, sizeof(*nl)) == -1) {
 		close(fd);
@@ -294,7 +279,7 @@ if_openlinksocket(void)
 	snl.nl_groups |= RTMGRP_IPV6_ROUTE | RTMGRP_IPV6_IFADDR | RTMGRP_NEIGH;
 #endif
 
-	return _open_link_socket(&snl, 1, NETLINK_ROUTE);
+	return _open_link_socket(&snl, NETLINK_ROUTE);
 }
 
 static int
@@ -437,7 +422,25 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 		case RTA_PRIORITY:
 			rt->metric = *(unsigned int *)RTA_DATA(rta);
 			break;
+		case RTA_METRICS:
+		{
+			struct rtattr *r2;
+			size_t l2;
+
+			l2 = rta->rta_len;
+			r2 = (struct rtattr *)RTA_DATA(rta);
+			while (RTA_OK(r2, l2)) {
+				switch (r2->rta_type) {
+				case RTAX_MTU:
+					rt->mtu = *(unsigned int *)RTA_DATA(r2);
+					break;
+				}
+				r2 = RTA_NEXT(r2, l2);
+			}
+			break;
 		}
+		}
+
 		rta = RTA_NEXT(rta, len);
 	}
 
@@ -498,6 +501,23 @@ if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
 		case RTA_PRIORITY:
 			rt->metric = *(unsigned int *)RTA_DATA(rta);
 			break;
+		case RTA_METRICS:
+		{
+			struct rtattr *r2;
+			size_t l2;
+
+			l2 = rta->rta_len;
+			r2 = (struct rtattr *)RTA_DATA(rta);
+			while (RTA_OK(r2, l2)) {
+				switch (r2->rta_type) {
+				case RTAX_MTU:
+					rt->mtu = *(unsigned int *)RTA_DATA(r2);
+					break;
+				}
+				r2 = RTA_NEXT(r2, l2);
+			}
+			break;
+		}
 		}
 		rta = RTA_NEXT(rta, len);
 	}
@@ -595,6 +615,14 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		errno = EBADMSG;
 		return -1;
 	}
+
+	/* Ignore messages generated by us.
+	 * For some reason we get messages generated by us
+	 * with a very large value in nlmsg_pid that seems to be
+	 * sequentially changing. Is there a better test for this? */
+	if (nlm->nlmsg_pid > get_max_pid_t())
+		return 1;
+
 	ifa = NLMSG_DATA(nlm);
 	if ((ifp = if_findindex(ctx->ifaces, ifa->ifa_index)) == NULL) {
 		/* We don't know about the interface the address is for
@@ -846,7 +874,7 @@ send_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 	static unsigned int seq;
 
 	memset(&snl, 0, sizeof(snl));
-	if ((s = _open_link_socket(&snl, 0, protocol)) == -1)
+	if ((s = _open_link_socket(&snl, protocol)) == -1)
 		return -1;
 	memset(&iov, 0, sizeof(iov));
 	iov.iov_base = hdr;
@@ -914,6 +942,27 @@ add_attr_32(struct nlmsghdr *n, unsigned short maxlen, unsigned short type,
 	return 0;
 }
 
+static int
+rta_add_attr_32(struct rtattr *rta, unsigned short maxlen,
+    unsigned short type, uint32_t data)
+{
+	unsigned short len = RTA_LENGTH(sizeof(data));
+	struct rtattr *subrta;
+
+	if (RTA_ALIGN(rta->rta_len) + len > maxlen) {
+		errno = ENOBUFS;
+		return -1;
+	}
+
+	subrta = (struct rtattr*)(void *)
+	    (((char*)rta) + RTA_ALIGN(rta->rta_len));
+	subrta->rta_type = type;
+	subrta->rta_len = len;
+	memcpy(RTA_DATA(subrta), &data, sizeof(data));
+	rta->rta_len = (unsigned short)(NLMSG_ALIGN(rta->rta_len) + len);
+	return 0;
+}
+
 #ifdef HAVE_NL80211_H
 static struct nlattr *
 nla_next(struct nlattr *nla, size_t *rem)
@@ -1137,28 +1186,10 @@ if_openrawsocket(struct interface *ifp, uint16_t protocol)
 	int n;
 #endif
 
-#ifdef SOCK_CLOEXEC
-	if ((s = socket(PF_PACKET, SOCK_DGRAM | SOCK_CLOEXEC | SOCK_NONBLOCK,
-	    htons(protocol))) == -1)
+	if ((s = xsocket(PF_PACKET, SOCK_DGRAM, htons(protocol),
+	    O_CLOEXEC | O_NONBLOCK)) == -1)
 		return -1;
-#else
-	int flags;
 
-	if ((s = socket(PF_PACKET, SOCK_DGRAM, htons(protocol))) == -1)
-		return -1;
-	if ((flags = fcntl(s, F_GETFD, 0)) == -1 ||
-	    fcntl(s, F_SETFD, flags | FD_CLOEXEC) == -1)
-	{
-		close(s);
-	        return -1;
-	}
-	if ((flags = fcntl(s, F_GETFL, 0)) == -1 ||
-	    fcntl(s, F_SETFL, flags | O_NONBLOCK) == -1)
-	{
-		close(s);
-	        return -1;
-	}
-#endif
 	/* Install the DHCP filter */
 	memset(&pf, 0, sizeof(pf));
 	if (protocol == ETHERTYPE_ARP) {
@@ -1230,7 +1261,6 @@ if_readrawpacket(struct interface *ifp, uint16_t protocol,
 		.msg_iov = &iov,
 		.msg_iovlen = 1,
 	};
-	struct dhcp_state *state;
 #ifdef PACKET_AUXDATA
 	unsigned char cmsgbuf[CMSG_LEN(sizeof(struct tpacket_auxdata))];
 	struct cmsghdr *cmsg;
@@ -1245,7 +1275,6 @@ if_readrawpacket(struct interface *ifp, uint16_t protocol,
 	msg.msg_controllen = sizeof(cmsgbuf);
 #endif
 
-	state = D_STATE(ifp);
 	fd = ipv4_protocol_fd(ifp, protocol);
 	bytes = recvmsg(fd, &msg, 0);
 	if (bytes == -1)
@@ -1307,7 +1336,8 @@ if_route(unsigned char cmd, const struct rt *rt)
 {
 	struct nlmr nlm;
 	int retval = 0;
-	struct dhcp_state *state;
+	const struct dhcp_state *state;
+	const struct ipv4ll_state *istate;
 
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
@@ -1328,14 +1358,18 @@ if_route(unsigned char cmd, const struct rt *rt)
 	nlm.rt.rtm_family = AF_INET;
 	nlm.rt.rtm_table = RT_TABLE_MAIN;
 
-	state = D_STATE(rt->iface);
+	state = D_CSTATE(rt->iface);
+	istate = IPV4LL_CSTATE(rt->iface);
 	if (cmd == RTM_DELETE)
 		nlm.rt.rtm_scope = RT_SCOPE_NOWHERE;
 	else {
 		/* We only change route metrics for kernel routes */
-		if (rt->dest.s_addr ==
+		if ((rt->dest.s_addr ==
 		    (state->addr.s_addr & state->net.s_addr) &&
-		    rt->net.s_addr == state->net.s_addr)
+		    rt->net.s_addr == state->net.s_addr) ||
+		    (istate && rt->dest.s_addr ==
+		    (istate->addr.s_addr & inaddr_llmask.s_addr) &&
+		    rt->net.s_addr == inaddr_llmask.s_addr))
 			nlm.rt.rtm_protocol = RTPROT_KERNEL;
 		else
 			nlm.rt.rtm_protocol = RTPROT_BOOT;
@@ -1355,7 +1389,8 @@ if_route(unsigned char cmd, const struct rt *rt)
 	    &rt->dest.s_addr, sizeof(rt->dest.s_addr));
 	if (nlm.rt.rtm_protocol == RTPROT_KERNEL) {
 		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_PREFSRC,
-		    &state->addr.s_addr, sizeof(state->addr.s_addr));
+		    istate == NULL ? &state->addr.s_addr : &istate->addr.s_addr,
+		    sizeof(state->addr.s_addr));
 	}
 	/* If a host route then don't add the gateway */
 	if ((cmd == RTM_ADD || cmd == RTM_CHANGE) &&
@@ -1368,6 +1403,17 @@ if_route(unsigned char cmd, const struct rt *rt)
 	if (rt->metric)
 		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY, rt->metric);
 
+	if (cmd != RTM_DELETE && rt->mtu) {
+		char metricsbuf[32];
+		struct rtattr *metrics = (void *)metricsbuf;
+
+		metrics->rta_type = RTA_METRICS;
+		metrics->rta_len = RTA_LENGTH(0);
+		rta_add_attr_32(metrics, sizeof(metricsbuf), RTAX_MTU, rt->mtu);
+		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_METRICS,
+		    RTA_DATA(metrics), (unsigned short)RTA_PAYLOAD(metrics));
+	}
+
 	if (send_netlink(rt->iface->ctx, NULL,
 	    NETLINK_ROUTE, &nlm.hdr, NULL) == -1)
 		retval = -1;
@@ -1482,27 +1528,6 @@ if_address6(const struct ipv6_addr *ap, int action)
 	return retval;
 }
 
-static int
-rta_add_attr_32(struct rtattr *rta, unsigned short maxlen,
-    unsigned short type, uint32_t data)
-{
-	unsigned short len = RTA_LENGTH(sizeof(data));
-	struct rtattr *subrta;
-
-	if (RTA_ALIGN(rta->rta_len) + len > maxlen) {
-		errno = ENOBUFS;
-		return -1;
-	}
-
-	subrta = (struct rtattr*)(void *)
-	    (((char*)rta) + RTA_ALIGN(rta->rta_len));
-	subrta->rta_type = type;
-	subrta->rta_len = len;
-	memcpy(RTA_DATA(subrta), &data, sizeof(data));
-	rta->rta_len = (unsigned short)(NLMSG_ALIGN(rta->rta_len) + len);
-	return 0;
-}
-
 int
 if_route6(unsigned char cmd, const struct rt6 *rt)
 {
@@ -1559,7 +1584,7 @@ if_route6(unsigned char cmd, const struct rt6 *rt)
 			add_attr_32(&nlm.hdr, sizeof(nlm),
 			    RTA_PRIORITY, rt->metric);
 	}
-	if (cmd == RTM_ADD && rt->mtu) {
+	if (cmd != RTM_DELETE && rt->mtu) {
 		char metricsbuf[32];
 		struct rtattr *metrics = (void *)metricsbuf;
 
