@@ -28,10 +28,13 @@
 #include <sys/socket.h>
 #include <sys/types.h>
 
-#include <netinet/in.h>
 #include <arpa/inet.h>
+#include <net/if.h>
 #include <net/route.h>
+#include <netinet/if_ether.h>
+#include <netinet/in.h>
 
+#include <assert.h>
 #include <ctype.h>
 #include <errno.h>
 #include <stdlib.h>
@@ -46,6 +49,7 @@
 #include "if.h"
 #include "if-options.h"
 #include "ipv4.h"
+#include "ipv4ll.h"
 #include "script.h"
 
 #define IPV4_LOOPBACK_ROUTE
@@ -156,24 +160,17 @@ ipv4_findaddr(struct dhcpcd_ctx *ctx, const struct in_addr *addr)
 }
 
 int
-ipv4_addrexists(struct dhcpcd_ctx *ctx, const struct in_addr *addr)
+ipv4_hasaddr(const struct interface *ifp)
 {
-	struct interface *ifp;
-	struct dhcp_state *state;
+	const struct dhcp_state *dstate;
+	const struct ipv4ll_state *istate;
 
-	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
-		state = D_STATE(ifp);
-		if (state) {
-			if (addr == NULL) {
-				if (state->addr.s_addr != INADDR_ANY)
-					return 1;
-			} else if (addr->s_addr == state->addr.s_addr)
-				return 1;
-		}
-		if (addr != NULL && ipv4_iffindaddr(ifp, addr, NULL))
-			return 1;
-	}
-	return 0;
+	dstate = D_CSTATE(ifp);
+	istate = IPV4LL_CSTATE(ifp);
+	return ((dstate &&
+	    dstate->added == STATE_ADDED &&
+	    dstate->addr.s_addr != INADDR_ANY) ||
+	    (istate && istate->addr.s_addr != INADDR_ANY));
 }
 
 void
@@ -205,6 +202,25 @@ ipv4_init(struct dhcpcd_ctx *ctx)
 	return 0;
 }
 
+int
+ipv4_protocol_fd(const struct interface *ifp, uint16_t protocol)
+{
+
+	if (protocol == ETHERTYPE_ARP) {
+		const struct iarp_state *istate;
+
+		istate = ARP_CSTATE(ifp);
+		assert(istate != NULL);
+		return istate->fd;
+	} else {
+		const struct dhcp_state *dstate;
+
+		dstate = D_CSTATE(ifp);
+		assert(dstate != NULL);
+		return dstate->raw_fd;
+	}
+}
+
 /* Interface comparer for working out ordering. */
 int
 ipv4_ifcmp(const struct interface *si, const struct interface *ti)
@@ -277,6 +293,11 @@ desc_route(const char *cmd, const struct rt *rt)
 	else if (rt->net.s_addr == htonl(INADDR_BROADCAST))
 		logger(ctx, LOG_INFO, "%s: %s host route to %s via %s",
 		    ifname, cmd, addr, inet_ntoa(rt->gate));
+	else if (rt->dest.s_addr == htonl(INADDR_ANY) &&
+	    rt->net.s_addr == htonl(INADDR_ANY) &&
+	    rt->gate.s_addr == htonl(INADDR_ANY))
+		logger(ctx, LOG_INFO, "%s: %s default route",
+		    ifname, cmd);
 	else if (rt->gate.s_addr == htonl(INADDR_ANY))
 		logger(ctx, LOG_INFO, "%s: %s route to %s/%d",
 		    ifname, cmd, addr, inet_ntocidr(rt->net));
@@ -364,6 +385,7 @@ ipv4_handlert(struct dhcpcd_ctx *ctx, int cmd, struct rt *rt)
 static int
 nc_route(struct rt *ort, struct rt *nrt)
 {
+	int change;
 
 	/* Don't set default routes if not asked to */
 	if (nrt->dest.s_addr == 0 &&
@@ -373,6 +395,7 @@ nc_route(struct rt *ort, struct rt *nrt)
 
 	desc_route(ort == NULL ? "adding" : "changing", nrt);
 
+	change = 0;
 	if (ort == NULL) {
 		ort = ipv4_findrt(nrt->iface->ctx, nrt, 0);
 		if (ort &&
@@ -382,8 +405,12 @@ nc_route(struct rt *ort, struct rt *nrt)
 		    ort->metric == nrt->metric &&
 #endif
 		    ort->gate.s_addr == nrt->gate.s_addr)))
-			return 0;
-	} else if (ort->flags & STATE_FAKE && !(nrt->flags & STATE_FAKE) &&
+		{
+			if (ort->mtu == nrt->mtu)
+				return 0;
+			change = 1;
+		}
+	} else if (ort->state & STATE_FAKE && !(nrt->state & STATE_FAKE) &&
 	    ort->iface == nrt->iface &&
 #ifdef HAVE_ROUTE_METRIC
 	    ort->metric == nrt->metric &&
@@ -391,12 +418,32 @@ nc_route(struct rt *ort, struct rt *nrt)
 	    ort->dest.s_addr == nrt->dest.s_addr &&
 	    ort->net.s_addr ==  nrt->net.s_addr &&
 	    ort->gate.s_addr == nrt->gate.s_addr)
-		return 0;
+	{
+		if (ort->mtu == nrt->mtu)
+			return 0;
+		change = 1;
+	}
+
+#ifdef RTF_CLONING
+	/* BSD can set routes to be cloning routes.
+	 * Cloned routes inherit the parent flags.
+	 * As such, we need to delete and re-add the route to flush children
+	 * to correct the flags. */
+	if (change && ort != NULL && ort->flags & RTF_CLONING)
+		change = 0;
+#endif
+
+	if (change) {
+		if (if_route(RTM_CHANGE, nrt) == 0)
+			return 0;
+		if (errno != ESRCH)
+			logger(nrt->iface->ctx, LOG_ERR, "if_route (CHG): %m");
+	}
 
 #ifdef HAVE_ROUTE_METRIC
 	/* With route metrics, we can safely add the new route before
 	 * deleting the old route. */
-	if (if_route(RTM_ADD, nrt)  == 0) {
+	if (if_route(RTM_ADD, nrt) == 0) {
 		if (ort && if_route(RTM_DELETE, ort) == -1 && errno != ESRCH)
 			logger(nrt->iface->ctx, LOG_ERR, "if_route (DEL): %m");
 		return 0;
@@ -453,7 +500,7 @@ add_subnet_route(struct rt_head *rt, const struct interface *ifp)
 		return rt;
 #endif
 
-	if ((r = malloc(sizeof(*r))) == NULL) {
+	if ((r = calloc(1, sizeof(*r))) == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
 		ipv4_freeroutes(rt);
 		return NULL;
@@ -461,6 +508,8 @@ add_subnet_route(struct rt_head *rt, const struct interface *ifp)
 	r->dest.s_addr = s->addr.s_addr & s->net.s_addr;
 	r->net.s_addr = s->net.s_addr;
 	r->gate.s_addr = INADDR_ANY;
+	r->mtu = dhcp_get_mtu(ifp);
+	r->src = s->addr;
 
 	TAILQ_INSERT_HEAD(rt, r, next);
 	return rt;
@@ -480,15 +529,16 @@ add_loopback_route(struct rt_head *rt, const struct interface *ifp)
 	if (s->addr.s_addr == INADDR_ANY)
 		return rt;
 
-	r = malloc(sizeof(*r));
-	if (r == NULL) {
+	if ((r = calloc(1, sizeof(*r))) == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
 		ipv4_freeroutes(rt);
 		return NULL;
 	}
-	r->dest.s_addr = s->addr.s_addr;
+	r->dest = s->addr;
 	r->net.s_addr = INADDR_BROADCAST;
 	r->gate.s_addr = htonl(INADDR_LOOPBACK);
+	r->mtu = dhcp_get_mtu(ifp);
+	r->src = s->addr;
 	TAILQ_INSERT_HEAD(rt, r, next);
 	return rt;
 }
@@ -499,15 +549,16 @@ get_routes(struct interface *ifp)
 {
 	struct rt_head *nrt;
 	struct rt *rt, *r = NULL;
+	const struct dhcp_state *state;
 
 	if (ifp->options->routes && TAILQ_FIRST(ifp->options->routes)) {
-		nrt = malloc(sizeof(*nrt));
+		if ((nrt = malloc(sizeof(*nrt))) == NULL)
+			return NULL;
 		TAILQ_INIT(nrt);
 		TAILQ_FOREACH(rt, ifp->options->routes, next) {
 			if (rt->gate.s_addr == 0)
 				break;
-			r = malloc(sizeof(*r));
-			if (r == NULL) {
+			if ((r = calloc(1, sizeof(*r))) == NULL) {
 				logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
 				ipv4_freeroutes(nrt);
 				return NULL;
@@ -515,53 +566,42 @@ get_routes(struct interface *ifp)
 			memcpy(r, rt, sizeof(*r));
 			TAILQ_INSERT_TAIL(nrt, r, next);
 		}
-		return nrt;
-	}
+	} else
+		nrt = dhcp_get_routes(ifp);
 
-	return get_option_routes(ifp, D_STATE(ifp)->new);
-}
-
-/* Some DHCP servers add set host routes by setting the gateway
- * to the assigned IP address or the destination address.
- * We need to change this. */
-static struct rt_head *
-massage_host_routes(struct rt_head *rt, const struct interface *ifp)
-{
-	struct rt *r;
-
-	if (rt) {
-		TAILQ_FOREACH(r, rt, next) {
-			if (r->gate.s_addr == D_CSTATE(ifp)->addr.s_addr ||
-			    r->gate.s_addr == r->dest.s_addr)
-			{
-				r->gate.s_addr = htonl(INADDR_ANY);
-				r->net.s_addr = htonl(INADDR_BROADCAST);
-			}
+	/* Copy our address as the source address */
+	if (nrt) {
+		state = D_CSTATE(ifp);
+		TAILQ_FOREACH(rt, nrt, next) {
+			rt->src = state->addr;
 		}
 	}
-	return rt;
-}
 
+	return nrt;
+}
 
 static struct rt_head *
 add_destination_route(struct rt_head *rt, const struct interface *ifp)
 {
 	struct rt *r;
+	const struct dhcp_state *state;
 
 	if (rt == NULL || /* failed a malloc earlier probably */
 	    !(ifp->flags & IFF_POINTOPOINT) ||
-	    !has_option_mask(ifp->options->dstmask, DHO_ROUTER))
+	    !has_option_mask(ifp->options->dstmask, DHO_ROUTER) ||
+	    (state = D_CSTATE(ifp)) == NULL)
 		return rt;
 
-	r = malloc(sizeof(*r));
-	if (r == NULL) {
+	if ((r = calloc(1, sizeof(*r))) == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
 		ipv4_freeroutes(rt);
 		return NULL;
 	}
 	r->dest.s_addr = INADDR_ANY;
 	r->net.s_addr = INADDR_ANY;
-	r->gate.s_addr = D_CSTATE(ifp)->dst.s_addr;
+	r->gate.s_addr = state->dst.s_addr;
+	r->mtu = dhcp_get_mtu(ifp);
+	r->src = state->addr;
 	TAILQ_INSERT_HEAD(rt, r, next);
 	return rt;
 }
@@ -603,7 +643,8 @@ add_router_host_route(struct rt_head *rt, const struct interface *ifp)
 		}
 		if (rtn != rtp)
 			continue;
-		state = D_CSTATE(ifp);
+		if ((state = D_CSTATE(ifp)) == NULL)
+			continue;
 		ifo = ifp->options;
 		if (ifp->flags & IFF_NOARP) {
 			if (!(ifo->options & DHCPCD_ROUTER_HOST_ROUTE_WARNED) &&
@@ -625,8 +666,7 @@ add_router_host_route(struct rt_head *rt, const struct interface *ifp)
 			    "%s: router %s requires a host route",
 			    ifp->name, inet_ntoa(rtp->gate));
 		}
-		rtn = malloc(sizeof(*rtn));
-		if (rtn == NULL) {
+		if ((rtn = calloc(1, sizeof(*rtn))) == NULL) {
 			logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
 			ipv4_freeroutes(rt);
 			return NULL;
@@ -634,37 +674,104 @@ add_router_host_route(struct rt_head *rt, const struct interface *ifp)
 		rtn->dest.s_addr = rtp->gate.s_addr;
 		rtn->net.s_addr = htonl(INADDR_BROADCAST);
 		rtn->gate.s_addr = htonl(INADDR_ANY);
+		rtn->mtu = dhcp_get_mtu(ifp);
+		rtn->src = state->addr;
 		TAILQ_INSERT_BEFORE(rtp, rtn, next);
 	}
 	return rt;
 }
 
+static int
+ipv4_doroute(struct rt *rt, struct rt_head *nrs)
+{
+	const struct dhcp_state *state;
+	struct rt *or;
+
+	state = D_CSTATE(rt->iface);
+	rt->state = state->added & STATE_FAKE;
+#ifdef HAVE_ROUTE_METRIC
+	rt->metric = rt->iface->metric;
+#endif
+	/* Is this route already in our table? */
+	if ((find_route(nrs, rt, NULL)) != NULL)
+		return 0;
+	/* Do we already manage it? */
+	if ((or = find_route(rt->iface->ctx->ipv4_routes, rt, NULL))) {
+		if (state->added & STATE_FAKE)
+			return 0;
+		if (or->state & STATE_FAKE ||
+		    or->iface != rt->iface ||
+#ifdef HAVE_ROUTE_METRIC
+		    rt->metric != or->metric ||
+#endif
+		    rt->src.s_addr != or->src.s_addr ||
+		    rt->gate.s_addr != or->gate.s_addr ||
+		    rt->mtu != or->mtu)
+		{
+			if (c_route(or, rt) != 0)
+				return 0;
+		}
+		TAILQ_REMOVE(rt->iface->ctx->ipv4_routes, or, next);
+		free(or);
+	} else {
+		if (state->added & STATE_FAKE) {
+			if ((or = ipv4_findrt(rt->iface->ctx, rt, 1)) == NULL)
+				return 0;
+			rt->iface = or->iface;
+			rt->gate.s_addr = or->gate.s_addr;
+#ifdef HAVE_ROUTE_METRIC
+			rt->metric = or->metric;
+#endif
+			rt->mtu = or->mtu;
+			rt->flags = or->flags;
+		} else {
+			if (n_route(rt) != 0)
+				return 0;
+		}
+	}
+	return 1;
+}
+
 void
 ipv4_buildroutes(struct dhcpcd_ctx *ctx)
 {
 	struct rt_head *nrs, *dnr;
-	struct rt *or, *rt, *rtn;
+	struct rt *rt, *rtn;
 	struct interface *ifp;
 	const struct dhcp_state *state;
+	int has_default;
 
 	/* We need to have the interfaces in the correct order to ensure
 	 * our routes are managed correctly. */
 	if_sortinterfaces(ctx);
 
-	nrs = malloc(sizeof(*nrs));
-	if (nrs == NULL) {
+	if ((nrs = malloc(sizeof(*nrs))) == NULL) {
 		logger(ctx, LOG_ERR, "%s: %m", __func__);
 		return;
 	}
 	TAILQ_INIT(nrs);
 
+	has_default = 0;
 	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
 		state = D_CSTATE(ifp);
-		if (state == NULL || state->new == NULL || !state->added)
+		if (state != NULL && state->new != NULL && state->added) {
+			dnr = get_routes(ifp);
+			dnr = add_subnet_route(dnr, ifp);
+		} else
+			dnr = NULL;
+		if ((rt = ipv4ll_subnet_route(ifp)) != NULL) {
+			if (dnr == NULL) {
+				if ((dnr = malloc(sizeof(*dnr))) == NULL) {
+					logger(ifp->ctx, LOG_ERR,
+					    "%s: malloc %m", __func__);
+					continue;
+				}
+				TAILQ_INIT(dnr);
+			}
+			TAILQ_INSERT_HEAD(dnr, rt, next);
+		}
+		if (dnr == NULL)
 			continue;
-		dnr = get_routes(ifp);
-		dnr = massage_host_routes(dnr, ifp);
-		dnr = add_subnet_route(dnr, ifp);
 #ifdef IPV4_LOOPBACK_ROUTE
 		dnr = add_loopback_route(dnr, ifp);
 #endif
@@ -672,53 +779,33 @@ ipv4_buildroutes(struct dhcpcd_ctx *ctx)
 			dnr = add_router_host_route(dnr, ifp);
 			dnr = add_destination_route(dnr, ifp);
 		}
-		if (dnr == NULL) /* failed to malloc all new routes */
+		if (dnr == NULL)
 			continue;
 		TAILQ_FOREACH_SAFE(rt, dnr, next, rtn) {
 			rt->iface = ifp;
-#ifdef HAVE_ROUTE_METRIC
-			rt->metric = ifp->metric;
-#endif
-			rt->flags = state->added & STATE_FAKE;
-			/* Is this route already in our table? */
-			if ((find_route(nrs, rt, NULL)) != NULL)
-				continue;
-			rt->src.s_addr = state->addr.s_addr;
-			/* Do we already manage it? */
-			if ((or = find_route(ctx->ipv4_routes, rt, NULL))) {
-				if (state->added & STATE_FAKE)
-					continue;
-				if (or->flags & STATE_FAKE ||
-				    or->iface != ifp ||
-#ifdef HAVE_ROUTE_METRIC
-				    rt->metric != or->metric ||
-#endif
-				    or->src.s_addr != state->addr.s_addr ||
-				    rt->gate.s_addr != or->gate.s_addr)
-				{
-					if (c_route(or, rt) != 0)
-						continue;
-				}
-				TAILQ_REMOVE(ctx->ipv4_routes, or, next);
-				free(or);
-			} else {
-				if (state->added & STATE_FAKE) {
-					or = ipv4_findrt(ctx, rt, 1);
-					if (or == NULL ||
-					    or->gate.s_addr != rt->gate.s_addr)
- 						continue;
-				} else {
-					if (n_route(rt) != 0)
-						continue;
-				}
+			if (ipv4_doroute(rt, nrs) == 1) {
+				TAILQ_REMOVE(dnr, rt, next);
+				TAILQ_INSERT_TAIL(nrs, rt, next);
+				if (rt->dest.s_addr == INADDR_ANY)
+					has_default = 1;
 			}
-			rt->flags |= STATE_ADDED;
-			TAILQ_REMOVE(dnr, rt, next);
-			TAILQ_INSERT_TAIL(nrs, rt, next);
 		}
 		ipv4_freeroutes(dnr);
 	}
 
+	/* If we don't manage a default route, grab one without a
+	 * gateway for any IPv4LL enabled interfaces. */
+	if (!has_default) {
+		TAILQ_FOREACH(ifp, ctx->ifaces, next) {
+			if ((rt = ipv4ll_default_route(ifp)) != NULL) {
+				if (ipv4_doroute(rt, nrs) == 1)
+					TAILQ_INSERT_TAIL(nrs, rt, next);
+				else
+					free(rt);
+			}
+		}
+	}
+
 	/* Remove old routes we used to manage */
 	if (ctx->ipv4_routes) {
 		TAILQ_FOREACH(rt, ctx->ipv4_routes, next) {
@@ -735,12 +822,13 @@ ipv4_buildroutes(struct dhcpcd_ctx *ctx)
 
 int
 ipv4_deladdr(struct interface *ifp,
-    const struct in_addr *addr, const struct in_addr *net)
+    const struct in_addr *addr, const struct in_addr *net, int keeparp)
 {
 	struct dhcp_state *dstate;
 	int r;
 	struct ipv4_state *state;
 	struct ipv4_addr *ap;
+	struct arp_state *astate;
 
 	logger(ifp->ctx, LOG_DEBUG, "%s: deleting IP address %s/%d",
 	    ifp->name, inet_ntoa(*addr), inet_ntocidr(*net));
@@ -750,14 +838,8 @@ ipv4_deladdr(struct interface *ifp,
 	    errno != ENODEV)
 		logger(ifp->ctx, LOG_ERR, "%s: %s: %m", ifp->name, __func__);
 
-	dstate = D_STATE(ifp);
-	if (dstate->addr.s_addr == addr->s_addr &&
-	    dstate->net.s_addr == net->s_addr)
-	{
-		dstate->added = 0;
-		dstate->addr.s_addr = 0;
-		dstate->net.s_addr = 0;
-	}
+	if (!keeparp && (astate = arp_find(ifp, addr)) != NULL)
+		arp_free(astate);
 
 	state = IPV4_STATE(ifp);
 	TAILQ_FOREACH(ap, &state->addrs, next) {
@@ -769,6 +851,18 @@ ipv4_deladdr(struct interface *ifp,
 			break;
 		}
 	}
+
+	/* Have to do this last incase the function arguments
+	 * were these very pointers. */
+	dstate = D_STATE(ifp);
+	if (dstate &&
+	    dstate->addr.s_addr == addr->s_addr &&
+	    dstate->net.s_addr == net->s_addr)
+	{
+		dstate->added = 0;
+		dstate->addr.s_addr = 0;
+		dstate->net.s_addr = 0;
+	}
 	return r;
 }
 
@@ -784,11 +878,11 @@ delete_address(struct interface *ifp)
 	if (ifo->options & DHCPCD_INFORM ||
 	    (ifo->options & DHCPCD_STATIC && ifo->req_addr.s_addr == 0))
 		return 0;
-	r = ipv4_deladdr(ifp, &state->addr, &state->net);
+	r = ipv4_deladdr(ifp, &state->addr, &state->net, 0);
 	return r;
 }
 
-static struct ipv4_state *
+struct ipv4_state *
 ipv4_getstate(struct interface *ifp)
 {
 	struct ipv4_state *state;
@@ -803,89 +897,112 @@ ipv4_getstate(struct interface *ifp)
 		}
 		TAILQ_INIT(&state->addrs);
 		TAILQ_INIT(&state->routes);
+#ifdef BSD
+		state->buffer_size = state->buffer_len = state->buffer_pos = 0;
+		state->buffer = NULL;
+#endif
 	}
 	return state;
 }
 
-static int
-ipv4_addaddr(struct interface *ifp, const struct dhcp_lease *lease)
+struct ipv4_addr *
+ipv4_addaddr(struct interface *ifp, const struct in_addr *addr,
+    const struct in_addr *mask, const struct in_addr *bcast)
 {
 	struct ipv4_state *state;
 	struct ipv4_addr *ia;
 
 	if ((state = ipv4_getstate(ifp)) == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: ipv4_getstate: %m", __func__);
-		return -1;
+		return NULL;
 	}
 	if (ifp->options->options & DHCPCD_NOALIAS) {
 		struct ipv4_addr *ian;
 
 		TAILQ_FOREACH_SAFE(ia, &state->addrs, next, ian) {
-			if (ia->addr.s_addr != lease->addr.s_addr)
-				ipv4_deladdr(ifp, &ia->addr, &ia->net);
+			if (ia->addr.s_addr != addr->s_addr)
+				ipv4_deladdr(ifp, &ia->addr, &ia->net, 0);
 		}
 	}
 
 	if ((ia = malloc(sizeof(*ia))) == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
-		return -1;
+		return NULL;
 	}
 
 	logger(ifp->ctx, LOG_DEBUG, "%s: adding IP address %s/%d",
-	    ifp->name, inet_ntoa(lease->addr),
-	    inet_ntocidr(lease->net));
-	if (if_addaddress(ifp, &lease->addr, &lease->net, &lease->brd) == -1) {
+	    ifp->name, inet_ntoa(*addr), inet_ntocidr(*mask));
+	if (if_addaddress(ifp, addr, mask, bcast) == -1) {
 		if (errno != EEXIST)
 			logger(ifp->ctx, LOG_ERR, "%s: if_addaddress: %m",
 			    __func__);
 		free(ia);
-		return -1;
+		return NULL;
 	}
 
 	ia->iface = ifp;
-	ia->addr = lease->addr;
-	ia->net = lease->net;
+	ia->addr = *addr;
+	ia->net = *mask;
 #ifdef IN_IFF_TENTATIVE
 	ia->addr_flags = IN_IFF_TENTATIVE;
 #endif
 	TAILQ_INSERT_TAIL(&state->addrs, ia, next);
+	return ia;
+}
+
+static int
+ipv4_daddaddr(struct interface *ifp, const struct dhcp_lease *lease)
+{
+	struct dhcp_state *state;
+
+	if (ipv4_addaddr(ifp, &lease->addr, &lease->net, &lease->brd) == NULL)
+		return -1;
+
+	state = D_STATE(ifp);
+	state->added = STATE_ADDED;
+
+	state->addr.s_addr = lease->addr.s_addr;
+	state->net.s_addr = lease->net.s_addr;
+
 	return 0;
 }
 
-static void
-ipv4_finalisert(struct interface *ifp)
+int
+ipv4_preferanother(struct interface *ifp)
 {
-	const struct dhcp_state *state = D_CSTATE(ifp);
+	struct dhcp_state *state = D_STATE(ifp), *nstate;
+	struct interface *ifn;
+	int preferred;
 
-	/* Find any freshly added routes, such as the subnet route.
-	 * We do this because we cannot rely on recieving the kernel
-	 * notification right now via our link socket. */
-	if_initrt(ifp);
+	if (state == NULL)
+		return 0;
+
+	preferred = 0;
+	if (!state->added)
+		goto out;
+
+	TAILQ_FOREACH(ifn, ifp->ctx->ifaces, next) {
+		if (ifn == ifp)
+			break; /* We are already the most preferred */
+		nstate = D_STATE(ifn);
+		if (nstate && !nstate->added &&
+		    nstate->lease.addr.s_addr == state->addr.s_addr)
+		{
+			preferred = 1;
+			delete_address(ifp);
+			if (ifn->options->options & DHCPCD_ARP)
+				dhcp_bind(ifn);
+			else {
+				ipv4_daddaddr(ifn, &nstate->lease);
+				nstate->added = STATE_ADDED;
+			}
+			break;
+		}
+	}
+
+out:
 	ipv4_buildroutes(ifp->ctx);
-	script_runreason(ifp, state->reason);
-
-	dhcpcd_daemonise(ifp->ctx);
-}
-
-void
-ipv4_finaliseaddr(struct interface *ifp)
-{
-	struct dhcp_state *state = D_STATE(ifp);
-	struct dhcp_lease *lease;
-
-	lease = &state->lease;
-
-	/* Delete the old address if different */
-	if (state->addr.s_addr != lease->addr.s_addr &&
-	    state->addr.s_addr != 0 &&
-	    ipv4_iffindaddr(ifp, &lease->addr, NULL))
-		delete_address(ifp);
-
-	state->added = STATE_ADDED;
-	state->defend = 0;
-	state->addr.s_addr = lease->addr.s_addr;
-	state->net.s_addr = lease->net.s_addr;
-	ipv4_finalisert(ifp);
+	return preferred;
 }
 
 void
@@ -904,38 +1021,15 @@ ipv4_applyaddr(void *arg)
 	dhcp = state->new;
 	lease = &state->lease;
 
+	if_sortinterfaces(ifp->ctx);
 	if (dhcp == NULL) {
 		if ((ifo->options & (DHCPCD_EXITING | DHCPCD_PERSISTENT)) !=
 		    (DHCPCD_EXITING | DHCPCD_PERSISTENT))
 		{
-			if (state->added) {
-				struct in_addr addr;
-
-				addr = state->addr;
+			if (state->added && !ipv4_preferanother(ifp)) {
 				delete_address(ifp);
-				TAILQ_FOREACH(ifn, ifp->ctx->ifaces, next) {
-					if (ifn == ifp ||
-					    strcmp(ifn->name, ifp->name) == 0)
-						continue;
-					nstate = D_STATE(ifn);
-					if (nstate && !nstate->added &&
-					    nstate->addr.s_addr == addr.s_addr)
-					{
-						if (ifn->options->options &
-						    DHCPCD_ARP)
-						{
-							dhcp_bind(ifn, NULL);
-						} else {
-							ipv4_addaddr(ifn,
-							    &nstate->lease);
-							nstate->added =
-							    STATE_ADDED;
-						}
-						break;
-					}
-				}
+				ipv4_buildroutes(ifp->ctx);
 			}
-			ipv4_buildroutes(ifp->ctx);
 			script_runreason(ifp, state->reason);
 		} else
 			ipv4_buildroutes(ifp->ctx);
@@ -943,29 +1037,29 @@ ipv4_applyaddr(void *arg)
 	}
 
 	/* Ensure only one interface has the address */
+	r = 0;
 	TAILQ_FOREACH(ifn, ifp->ctx->ifaces, next) {
-		if (ifn == ifp || strcmp(ifn->name, ifp->name) == 0)
+		if (ifn == ifp) {
+			r = 1; /* past ourselves */
 			continue;
+		}
 		nstate = D_STATE(ifn);
 		if (nstate && nstate->added &&
 		    nstate->addr.s_addr == lease->addr.s_addr)
 		{
-			if (ifn->metric <= ifp->metric) {
+			if (r == 0) {
 				logger(ifp->ctx, LOG_INFO,
 				    "%s: preferring %s on %s",
 				    ifp->name,
 				    inet_ntoa(lease->addr),
 				    ifn->name);
-				state->addr.s_addr = lease->addr.s_addr;
-				state->net.s_addr = lease->net.s_addr;
-				ipv4_finalisert(ifp);
+				return;
 			}
 			logger(ifp->ctx, LOG_INFO, "%s: preferring %s on %s",
 			    ifn->name,
 			    inet_ntoa(lease->addr),
 			    ifp->name);
-			ipv4_deladdr(ifn, &nstate->addr, &nstate->net);
-			nstate->added = 0;
+			ipv4_deladdr(ifn, &nstate->addr, &nstate->net, 0);
 			break;
 		}
 	}
@@ -973,18 +1067,18 @@ ipv4_applyaddr(void *arg)
 	/* Does another interface already have the address from a prior boot? */
 	if (ifn == NULL) {
 		TAILQ_FOREACH(ifn, ifp->ctx->ifaces, next) {
-			if (ifn == ifp || strcmp(ifn->name, ifp->name) == 0)
+			if (ifn == ifp)
 				continue;
 			ap = ipv4_iffindaddr(ifn, &lease->addr, NULL);
 			if (ap)
-				ipv4_deladdr(ifn, &ap->addr, &ap->net);
+				ipv4_deladdr(ifn, &ap->addr, &ap->net, 0);
 		}
 	}
 
 	/* If the netmask is different, delete the addresss */
 	ap = ipv4_iffindaddr(ifp, &lease->addr, NULL);
 	if (ap && ap->net.s_addr != lease->net.s_addr)
-		ipv4_deladdr(ifp, &ap->addr, &ap->net);
+		ipv4_deladdr(ifp, &ap->addr, &ap->net, 0);
 
 	if (ipv4_iffindaddr(ifp, &lease->addr, &lease->net))
 		logger(ifp->ctx, LOG_DEBUG,
@@ -992,7 +1086,7 @@ ipv4_applyaddr(void *arg)
 		    ifp->name, inet_ntoa(lease->addr),
 		    inet_ntocidr(lease->net));
 	else {
-		r = ipv4_addaddr(ifp, lease);
+		r = ipv4_daddaddr(ifp, lease);
 		if (r == -1 && errno != EEXIST)
 			return;
 	}
@@ -1007,7 +1101,24 @@ ipv4_applyaddr(void *arg)
 		return;
 #endif
 
-	ipv4_finaliseaddr(ifp);
+	/* Delete the old address if different */
+	if (state->addr.s_addr != lease->addr.s_addr &&
+	    state->addr.s_addr != 0 &&
+	    ipv4_iffindaddr(ifp, &lease->addr, NULL))
+		delete_address(ifp);
+
+	state->added = STATE_ADDED;
+	state->addr.s_addr = lease->addr.s_addr;
+	state->net.s_addr = lease->net.s_addr;
+
+	/* Find any freshly added routes, such as the subnet route.
+	 * We do this because we cannot rely on recieving the kernel
+	 * notification right now via our link socket. */
+	if_initrt(ifp);
+	ipv4_buildroutes(ifp->ctx);
+	script_runreason(ifp, state->reason);
+
+	dhcpcd_daemonise(ifp->ctx);
 }
 
 void
@@ -1079,6 +1190,9 @@ ipv4_free(struct interface *ifp)
 				free(addr);
 			}
 			ipv4_freerts(&state->routes);
+#ifdef BSD
+			free(state->buffer);
+#endif
 			free(state);
 		}
 	}
